""" Define custom exceptions """
from typing import Union, List

PrimaryAttribute = Union[List[str], str]


class InvalidSeason(Exception):
    """ Invalid season """

    def __init__(self, message: str, season: str) -> None:
        super().__init__(message)
        self.season = season


class InvalidPlayer(Exception):
    """ Invalid player """

    def __init__(self, message: str, player: str) -> None:
        super().__init__(message)
        self.player = player


class InvalidLeague(Exception):
    """ Invalid league """

    def __init__(self, message: str, league: str) -> None:
        super().__init__(message)
        self.league = league


class InvalidTeam(Exception):
    """ Invalid team """

    def __init__(self, message: str, team: str) -> None:
        super().__init__(message)
        self.team = team


class InvalidMatch(Exception):
    """ Invalid match """

    def __init__(self, message: str, match: str) -> None:
        super().__init__(message)
        self.match = match



""" understatAPI client """
from types import TracebackType
from typing import Iterator
import requests
from selenium.common.exceptions import WebDriverException
from .utils import get_public_methods, str_to_class, find_endpoints
from .endpoints import (
    LeagueEndpoint,
    PlayerEndpoint,
    TeamEndpoint,
    MatchEndpoint,
)
from .services import Search
from .exceptions import PrimaryAttribute


class UnderstatClient:
    """#pylint: disable=line-too-long
    API client for understat

    The main interface for interacting with understatAPI. Exposes
    each of the entrypoints and services, maintains a consistent
    session and handles errors

    :Example:

    .. code-block::

        from understatapi import UnderstatClient

        with UnderstatClient() as understat:
            league_player_data = understat.league(league="EPL").get_player_data(season="2019")
            player_shot_data = understat.player(player="2371").get_shot_data()
            team_match_data = understat.team(team="Manchester_United").get_match_data(season="2019")
            roster_data = understat.match(match="14711").get_roster_data()
            player_match_data = understat.search("Cristiano Ronaldo").get_match_data()

    Using the context manager gives some more verbose error handling

    .. testsetup::

            from understatapi import UnderstatClient

    .. doctest::

        >>> team=""
        >>> with UnderstatClient() as understat:
        ...     understat.team(team).get_bad_data() # doctest: +SKIP
        Traceback (most recent call last)
        File "<stdin>", line 2, in <module>
        File "understatapi/api.py", line 59, in __exit__
            raise AttributeError(
        AttributeError: 'TeamEndpoint' object has no attribute 'get_bad_data'
        Its public methods are ['get_context_data', 'get_match_data', 'get_player_data']

    """

    def __init__(self) -> None:
        self.session = requests.Session()

    def __enter__(self) -> "UnderstatClient":
        return self

    def __exit__(
        self,
        exception_type: type,
        exception_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        if exception_type is AttributeError:
            endpoint = find_endpoints(str(exception_value))
            endpoint_obj = str_to_class(__name__, endpoint[0])
            public_methods = get_public_methods(endpoint_obj)
            raise AttributeError(
                str(exception_value)
                + f"\nIts public methods are {public_methods}"
            )
        self.session.close()

    def league(self, league: PrimaryAttribute) -> LeagueEndpoint:
        """
        Endpoint for league data. Use this function to get data from a
        url of the form ``https://understat.com/league/<league>/<season>``

        :param league: Name of the league(s) to get data for,
            one of {EPL, La_Liga, Bundesliga, Serie_A, Ligue_1, RFPL}
        :rtype: :py:class:`~understatapi.endpoints.league.LeagueEndpoint`

        :Example:

        .. testsetup::

            from understatapi import UnderstatClient

        .. doctest::

            >>> leagues = ["EPL", "Bundesliga"]
            >>> with UnderstatClient() as understat:
            ...     for league in understat.league(leagues):
            ...         print(league.league)
            EPL
            Bundesliga

        """
        return LeagueEndpoint(league=league, session=self.session)

    def player(self, player: PrimaryAttribute) -> PlayerEndpoint:
        """
        Endpoint for player data. Use this function to get data from a
        url of the form ``https://understat.com/player/<player_id>/``

        :param player: Id of the player(s) to get data for
        :rtype: :py:class:`~understatapi.endpoints.player.PlayerEndpoint`

        :Example:

        .. testsetup::

            from understatapi import UnderstatClient

        .. doctest::

            >>> player_ids = ["000", "111"]
            >>> with UnderstatClient() as understat:
            ...     for player in understat.player(player_ids):
            ...         print(player.player)
            000
            111

        """
        return PlayerEndpoint(player=player, session=self.session)

    def team(self, team: PrimaryAttribute) -> TeamEndpoint:
        """
        Endpoint for team data. Use this function to get data from a
        url of the form ``https://understat.com/team/<team>/<season>``

        :param team: Name of the team(s) to get data for
        :rtype: :py:class:`~understatapi.endpoints.team.TeamEndpoint`

        :Example:

        .. testsetup::

            from understatapi import UnderstatClient

        .. doctest::

            >>> team_names = ["Manchester_United", "Liverpool"]
            >>> with UnderstatClient() as understat:
            ...     for team in understat.team(team_names):
            ...         print(team.team)
            Manchester_United
            Liverpool

        """
        return TeamEndpoint(team=team, session=self.session)

    def match(self, match: PrimaryAttribute) -> MatchEndpoint:
        """
        Endpoint for match data. Use this function to get data from a
        url of the form ``https://understat.com/match/<match_id>``

        :param match: Id of match(es) to get data for
        :rtype: :class:`~understatapi.endpoints.match.MatchEndpoint`

        :Example:

        .. testsetup::

            from understatapi import UnderstatClient

        .. doctest::

            >>> match_ids = ["123", "456"]
            >>> with UnderstatClient() as understat:
            ...     for match in understat.match(match_ids):
            ...         print(match.match)
            123
            456

        """
        return MatchEndpoint(
            match=match,
            session=self.session,
        )

    def search(
        self, player_name: str, max_ids: int = 5, page_load_timeout: int = 5
    ) -> Iterator[PlayerEndpoint]:
        """#pylint: disable=line-too-long
        Search for a player by name

        :param player_name: Player name to enter into the
            seach bar
        :param max_ids: The maximum number of player ids to return
        :param page_load_timeout: Number of seconds to wait for the page
            to load before raising a ``TimeoutError``
        :rtype: :class:`Iterator` [:class:`~understatapi.endpoints.player.PlayerEndpoint`]


        :Example:

        .. testsetup::

            from minimock import Mock
            from understatapi import UnderstatClient
            from understatapi.endpoints import PlayerEndpoint
            from understatapi.services import Search
            players = [[
                "2371",
                "2028",
                "7097",
                ],
                ["2371"]
            ]
            Search.get_player_ids = Mock(
                "Search.get_player_ids",
                returns_iter=players,
                tracker=None
            )

        .. doctest::

            >>> player_name = "Ronaldo"
            >>> with UnderstatClient() as understat:
            ...     for player in understat.search(player_name):
            ...         print(player.player)
            2371
            2028
            7097
            >>> with UnderstatClient() as understat:
            ...     for player in understat.search(player_name, max_ids=1):
            ...         print(player.player)
            2371

        """
        try:
            with Search(
                player_name=player_name,
                session=self.session,
                max_ids=max_ids,
                page_load_timeout=page_load_timeout,
            ) as search:
                for player in search.get_player_ids():
                    yield PlayerEndpoint(player=player, session=self.session)
        except WebDriverException as err:
            raise WebDriverException(
                "You must have 'geckodriver' installed to "
                "use UnderstatClient.search()"
            ) from err



""" Helper functions for formatting data """
import sys
from typing import List
import inspect
import re


def get_all_methods(cls: type) -> List[str]:
    """
    Get the names of all methods in a class, excluding
    methods decorated with ``@property``, ``@classmethod``, etc

    :param cls: The class to get the methods for
    :return: A list of the names of the methods
    """
    return [meth[0] for meth in inspect.getmembers(cls, inspect.isfunction)]


def get_public_methods(cls: type) -> List[str]:
    """
    Get the names of all public methods in a class

    :param cls: The class to get all public methods for
    :return: A list of the names of the public methods
    """
    methods = get_all_methods(cls)
    methods = [meth for meth in methods if not meth.startswith("_")]
    return methods


def find_endpoints(line: str) -> List[str]:
    """
    Find the name of a subclass of
    ``~understatapi.endpoints.base.BaseEndpoint`` in a string

    :param line: The string in which to search for the name of a
        ``~understatapi.endpoints.base.BaseEndpoint`` object
    """
    match = re.findall(r"\w+Endpoint", line)
    if match is None:
        return []  # pragma: no cover
    return match


def str_to_class(modulename: str, classname: str) -> type:
    """
    Get a class by using its name
    """
    return getattr(sys.modules[modulename], classname)



""" Player endpoint """
from typing import Dict, Any
import requests
from requests.exceptions import HTTPError
from .base import BaseEndpoint
from ..exceptions import InvalidPlayer, PrimaryAttribute


class PlayerEndpoint(BaseEndpoint):
    """
    Use this class to get data from a url of the form
    ``https://understat.com/player/<player_id>/``

    :Example:

    .. testsetup::

        import requests
        from understatapi.endpoints import PlayerEndpoint

    .. testcleanup::

        session.close()

    .. doctest::

        >>> session = requests.Session()
        >>> player_ids = ["000", "111"]
        >>> for player in PlayerEndpoint(player_ids, session=session):
        ...     print(player.player)
        000
        111

    """

    queries = ["matchesData", "shotsData", "groupsData"]

    def __init__(
        self, player: PrimaryAttribute, session: requests.Session
    ) -> None:
        """
        :param player: Id of the player(s) to get data for
        :param session: The current session
        """
        self._primary_attr = player
        super().__init__(primary_attr=self._primary_attr, session=session)

    @property
    def player(self) -> PrimaryAttribute:
        """ player id """
        return self._primary_attr

    def _get_data(self, query: str, **kwargs: str) -> Dict[str, Any]:
        """
        Get data on a per-player basis

        :param query: Identifies the type of data to get,
            one of {matchesData, shotsData, groupsData}
        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        if not isinstance(self.player, str):
            raise TypeError("``player`` must be a string")
        self._check_args()
        url = self.base_url + "player/" + self.player

        try:
            data = self._get_response(url=url, query=query, **kwargs)
        except HTTPError as err:
            raise InvalidPlayer(
                f"{self.player} is not a valid player or player id",
                player=self.player,
            ) from err

        return data

    def get_match_data(self, **kwargs: str) -> Dict[str, Any]:
        """
        Get match level data for a player

        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(query="matchesData", **kwargs)
        return data

    def get_shot_data(self, **kwargs: str) -> Dict[str, Any]:
        """
        Get shot level data for a player

        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(query="shotsData", **kwargs)
        return data

    def get_season_data(self, **kwargs: str) -> Dict[str, Any]:
        """
        Get season level data for a player

        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(query="groupsData", **kwargs)
        return data



""" Mock a selenium driver object """
import os
import textwrap
import time
from lxml import html


class MockWebDriverWait:
    """ Mock WebDriverWait """

    def __init__(self, driver, timeout, wait_time=0.1):
        self._driver = driver
        self._timeout = timeout
        self._wait_time = wait_time

    def until(self, message=""):
        """ until """
        if self._wait_time < self._timeout:
            time.sleep(self._wait_time)
        else:
            raise TimeoutError(message)


class MockWebElement:
    """ Mock a selenium web element """

    def __init__(self, element=None):
        self.element = element

    def click(self):
        """ click """
        class_value = self.element.classes._get_class_value()
        if class_value:
            return True
        raise ValueError

    def send_keys(self, keys):
        # pylint: disable=unused-argument
        """ send keys """
        if isinstance(self.element, html.InputElement):
            return True
        raise TypeError

    @property
    def text(self):
        """ text """
        elem_str = str(self.element.text_content())
        elem_str = textwrap.dedent(elem_str)
        elem_str = os.linesep.join([s for s in elem_str.splitlines() if s])
        return elem_str


class MockWebDriver:
    """ Mock a selenium web driver """

    def __init__(self, options=None):
        self.options = options
        self._content = None
        self.url = None
        self._current_url = None

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.quit()

    def quit(self):
        """ quit """

    @property
    def current_url(self):
        """ current url """
        return self._current_url

    def find_element_by_xpath(self, xpath):
        """ find element by xpath """
        tree = html.fromstring(self._content)
        res = tree.xpath(xpath)
        element = MockWebElement(element=res[0])
        self._current_url = res[0].text
        return element

    def get(self, url):
        """ get """
        self.url = url
        with open(url, encoding="utf-8") as file:
            content = file.read()
        self._content = content
        return content

    def back(self):
        """ back """

    def get_cookies(self):
        # pylint: disable=no-self-use
        """ get cookies """
        return [
            {"name": "PHPSESSID", "value": "00000"},
            {"name": "UID", "value": "11111"},
        ]



""" Use the search bar of understat.com """
from types import TracebackType
from typing import Tuple, Sequence, Iterator
import requests
from selenium.webdriver import Firefox
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC


class Search:
    """#pylint: disable=line-too-long
    Use the search bar on ``understat.com``

    All results matching the given player name will be returned

    :Example:

    .. testsetup::

        import requests
        from minimock import Mock
        from understatapi.endpoints import PlayerEndpoint
        from understatapi.services import Search
        players = [[
            "2371",
            "2028",
            "7097",
            ],
            ["2371"]
        ]
        Search.get_player_ids = Mock("Search.get_player_ids", returns_iter=players, tracker=None)

    .. testcleanup::

        session.close()

    .. doctest::

        >>> session = requests.Session()
        >>> player_name = "Ronaldo"
        >>> with Search(player_name, session=session) as search:
        ...     for player_id in search.get_player_ids():
        ...         print(player_id)
        2371
        2028
        7097
        >>> with Search(player_name, session=session, max_ids=1) as search:
        ...     for player_id in search.get_player_ids():
        ...         print(player_id)
        2371

    """

    opts = Options()
    opts.set_headless()
    url = "https://understat.com/"
    _web_driver = Firefox

    def __init__(
        self,
        player_name: str,
        session: requests.Session,
        max_ids: int = 5,
        page_load_timeout: int = 5,
    ) -> None:
        """
        :param player_name: Name of player to search for
        :param session: The current session
        :param max_ids: The maximum number of player ids to return
        :param page_load_timeout: Number of seconds to wait for the page
            to load before raising a ``TimeoutError``
        """
        self.player_name = player_name
        self.session = session
        self.max_ids = max_ids
        self.page_load_timeout = page_load_timeout
        self._initialise_browser()

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}" f"({self.player_name})>"
        )  # pragma: no cover

    def __enter__(self) -> "Search":
        return self

    def __exit__(
        self,
        exception_type: type,
        exception_value: BaseException,
        traceback: TracebackType,
    ) -> None:
        cookies = self.browser.get_cookies()
        for cookie in cookies:
            self.session.cookies.set(cookie["name"], cookie["value"])
        self.browser.quit()

    def _initialise_browser(self) -> None:
        """ Initialise the Firefox WebDriver object """
        self.browser = self._web_driver(options=self.opts)

    def get_player_ids(self) -> Iterator[str]:
        """
        Get the ids that are returned when you search for the given
        player name
        """
        self.browser.get(self.url)
        self._make_search()
        players = self._get_search_results()
        for i, player_id in enumerate(self._cycle_results(players)):
            yield player_id
            if i >= self.max_ids - 1:
                break

    @staticmethod
    def _get_player_id_from_url(url: str) -> str:
        """
        Get the player id from a url
        """
        id_number = url.split("/")[-1]
        return id_number

    def _cycle_results(
        self, results: Sequence[Tuple[str, str]]
    ) -> Iterator[str]:
        """
        Cycles through each of the search results, getting the url each time
        """
        for i, _ in enumerate(results):
            self.browser.find_element_by_xpath(
                f"//*[@id='header']/div/nav[2]/ul/"
                f"li[1]/span[1]/div/div/div[{i+1}]"
            ).click()
            player_id = self._get_player_id_from_url(self.browser.current_url)
            self.browser.back()
            self._make_search()
            yield player_id

    def _wait_for_page_load(self, **kwargs: int) -> None:
        """ Wait for the page loading screen to disappear """
        try:
            WebDriverWait(
                self.browser, self.page_load_timeout, **kwargs
            ).until(
                EC.invisibility_of_element_located(
                    (By.XPATH, "/html/body/div[1]/div[1]")
                )
            )
        except TimeoutError as err:  # pragma: no cover
            raise TimeoutError(err) from err

    def _make_search(self, **kwargs: int) -> None:
        """
        Make the search
        """
        # Wait for loading page to disappear
        self._wait_for_page_load()
        # Open the search bar
        player_search = self.browser.find_element_by_xpath(
            "//*[@id='header']/div/nav[2]/ul/li[1]"
        )
        player_search.click()
        # Enter the player name into the search bar
        search_form = self.browser.find_element_by_xpath(
            "//*[@id='header']/div/nav[2]/ul/li[1]/span[1]/input"
        )
        search_form.send_keys(self.player_name)
        # Click the search form to make all the results appear
        search_form.click()
        WebDriverWait(self.browser, self.page_load_timeout, **kwargs).until(
            EC.visibility_of_element_located(
                (By.XPATH, "//*[@id='header']/div/nav[2]/ul/li[1]/span[1]/div")
            )
        )

    def _get_search_results(self) -> Sequence[Tuple[str, str]]:
        """
        Get results from the search
        """
        players = self.browser.find_element_by_xpath(
            "//*[@id='header']/div/nav[2]/ul/li[1]/span[1]/div"
        )
        player_club_list = players.text.split("\n")
        players = zip(player_club_list[::2], player_club_list[1::2])

        return players



# pylint: disable=unused-argument
# pylint: disable=invalid-name
# pylint: disable=too-many-instance-attributes
# pylint: disable=undefined-loop-variable
"""Test understatapi"""
from typing import Dict
import unittest
from unittest.mock import patch
import json
from test import mocked_requests_get
from test.mock_selenium import MockWebDriver, MockWebDriverWait
import requests
from selenium.common.exceptions import WebDriverException
from understatapi import UnderstatClient
from understatapi.services import Search
from understatapi.endpoints import BaseEndpoint, PlayerEndpoint
from understatapi.exceptions import (
    InvalidMatch,
    InvalidPlayer,
    InvalidTeam,
    InvalidLeague,
    InvalidSeason,
)


def read_json(path: str) -> Dict:
    """Read json data"""
    with open(path, "r", encoding="utf-8") as fh:
        data = json.load(fh)
    return data


class EndpointBaseTestCase(unittest.TestCase):
    """Base class for all endpoint ``unittest.TestCase``` classes"""

    def setUp(self):
        self.understat = UnderstatClient()
        self.match_id = "dummy_match"
        self.match = self.understat.match(self.match_id)
        self.league_name = "EPL"
        self.league = self.understat.league(self.league_name)
        self.player_id = "dummy_player"
        self.player = self.understat.player(self.player_id)
        self.team_name = "dummy_team"
        self.team = self.understat.team(self.team_name)

    def tearDown(self):
        self.understat.session.close()


@patch.object(requests.Session, "get")
class TestEndpointsResponse(EndpointBaseTestCase):
    """Test that endpoints return the expected output"""

    def test_match_get_shot_data(self, mock_get):
        """ test ``match.get_shot_data()`` """
        mock_get.return_value = mocked_requests_get(
            "test/resources/match.html"
        )
        data = self.match.get_shot_data()
        data_path = "test/resources/data/match_shotsdata.json"
        expected_data = read_json(data_path)
        self.assertDictEqual(expected_data, data)

    def test_match_get_roster_data(self, mock_get):
        """ test ``match.get_roster_data()`` """
        mock_get.return_value = mocked_requests_get(
            "test/resources/match.html"
        )
        data = self.match.get_roster_data()
        data_path = "test/resources/data/match_rostersdata.json"
        expected_data = read_json(data_path)
        self.assertDictEqual(expected_data, data)

    def test_match_get_match_info(self, mock_get):
        """ test ``match.get_match_info()`` """
        mock_get.return_value = mocked_requests_get(
            "test/resources/match.html"
        )
        data = self.match.get_match_info()
        data_path = "test/resources/data/match_matchinfo.json"
        expected_data = read_json(data_path)
        self.assertDictEqual(expected_data, data)

    def test_player_get_match_data(self, mock_get):
        """ test ``player.get_match_data()`` """
        mock_get.return_value = mocked_requests_get(
            "test/resources/player.html"
        )
        data = self.player.get_match_data()
        data_path = "test/resources/data/player_matchesdata.json"
        expected_data = read_json(data_path)
        for i, (record, expected_record) in enumerate(
            zip(data, expected_data)
        ):
            with self.subTest(record=i):
                self.assertDictEqual(record, expected_record)

    def test_get_shot_data_return_value(self, mock_get):
        """ test ``player.get_shot_data()`` """
        mock_get.return_value = mocked_requests_get(
            "test/resources/player.html"
        )
        data = self.player.get_shot_data()
        data_path = "test/resources/data/player_shotsdata.json"
        expected_data = read_json(data_path)
        for i, (record, expected_record) in enumerate(
            zip(data, expected_data)
        ):
            with self.subTest(record=i):
                self.assertDictEqual(record, expected_record)

    def test_player_get_season_data(self, mock_get):
        """ test ``player.get_season_data()`` """
        mock_get.return_value = mocked_requests_get(
            "test/resources/player.html"
        )
        data = self.player.get_season_data()
        data_path = "test/resources/data/player_groupsdata.json"
        expected_data = read_json(data_path)
        self.assertDictEqual(data, expected_data)

    def test_team_get_player_data(self, mock_get):
        """ test ``team.get_match_data()`` """
        mock_get.return_value = mocked_requests_get("test/resources/team.html")
        data = self.team.get_player_data(season="2019")
        data_path = "test/resources/data/team_playersdata.json"
        expected_data = read_json(data_path)
        for i, (record, expected_record) in enumerate(
            zip(data, expected_data)
        ):
            with self.subTest(record=i):
                self.assertDictEqual(record, expected_record)

    def test_team_get_match_data(self, mock_get):
        """ test ``team.get_match_data()`` """
        mock_get.return_value = mocked_requests_get("test/resources/team.html")
        data = self.team.get_match_data(season="2019")
        data_path = "test/resources/data/team_datesdata.json"
        expected_data = read_json(data_path)
        for i, (record, expected_record) in enumerate(
            zip(data, expected_data)
        ):
            with self.subTest(record=i):
                self.assertDictEqual(record, expected_record)

    def test_team_get_context_data(self, mock_get):
        """ test ``team.get_context_data()`` """
        mock_get.return_value = mocked_requests_get("test/resources/team.html")
        data = self.team.get_context_data(season="2019")
        data_path = "test/resources/data/team_statisticsdata.json"
        expected_data = read_json(data_path)
        self.assertDictEqual(data, expected_data)

    def test_league_get_team_data(self, mock_get):
        """ test ``league.get_team_data()`` """
        mock_get.return_value = mocked_requests_get(
            "test/resources/league_epl.html"
        )
        data = self.league.get_team_data(season="2019")
        data_path = "test/resources/data/league_teamsdata.json"
        expected_data = read_json(data_path)
        self.assertDictEqual(data, expected_data)

    def test_league_get_match_data(self, mock_get):
        """ test ``league.get_match_data()`` """
        mock_get.return_value = mocked_requests_get(
            "test/resources/league_epl.html"
        )
        data = self.league.get_match_data(season="2019")
        data_path = "test/resources/data/league_datesdata.json"
        expected_data = read_json(data_path)
        for i, (record, expected_record) in enumerate(
            zip(data, expected_data)
        ):
            with self.subTest(record=i):
                self.assertDictEqual(record, expected_record)

    def test_league_get_player_data(self, mock_get):
        """ test ``league.get_player_data()`` """
        mock_get.return_value = mocked_requests_get(
            "test/resources/league_epl.html"
        )
        data = self.league.get_player_data(season="2019")
        data_path = "test/resources/data/league_playersdata.json"
        expected_data = read_json(data_path)
        for i, (record, expected_record) in enumerate(
            zip(data, expected_data)
        ):
            with self.subTest(record=i):
                self.assertDictEqual(record, expected_record)


@patch.object(BaseEndpoint, "_request_url")
@patch.object(requests.Session, "get", side_effect=mocked_requests_get)
@patch.object(BaseEndpoint, "_get_response")
class TestEndpointArguments(EndpointBaseTestCase):
    """Test that endpoints receive the expectred arguments"""

    def test_match_get_shot_data_args(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``match.get_shot_data()`` """
        self.match.get_shot_data()
        mock_get_response.assert_called_with(
            url=f"https://understat.com/match/{self.match_id}",
            query="shotsData",
        )

    def test_match_get_roster(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``match.get_roster_data()`` """
        self.match.get_roster_data()
        mock_get_response.assert_called_with(
            url=f"https://understat.com/match/{self.match_id}",
            query="rostersData",
        )

    def test_match_get_match_info(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``match.get_match_info()`` """
        self.match.get_match_info()
        mock_get_response.assert_called_with(
            url=f"https://understat.com/match/{self.match_id}",
            query="match_info",
        )

    def test_player_get_match_data(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``player.get_match_data()`` """
        self.player.get_match_data()
        mock_get_response.assert_called_with(
            url=f"https://understat.com/player/{self.player_id}",
            query="matchesData",
        )

    def test_player_get_shot_data(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``player.get_shot_data()`` """
        self.player.get_shot_data()
        mock_get_response.assert_called_with(
            url=f"https://understat.com/player/{self.player_id}",
            query="shotsData",
        )

    def test_player_get_season_data(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``player.get_season_data()`` """
        self.player.get_season_data()
        mock_get_response.assert_called_with(
            url=f"https://understat.com/player/{self.player_id}",
            query="groupsData",
        )

    def test_team_get_player_data(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``team.get_player_data()`` """
        self.team.get_player_data(season="2019")
        mock_get_response.assert_called_with(
            url=f"https://understat.com/team/{self.team_name}/2019",
            query="playersData",
        )

    def test_team_get_match_data(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``team.get_match_data()`` """
        self.team.get_match_data(season="2019")
        mock_get_response.assert_called_with(
            url=f"https://understat.com/team/{self.team_name}/2019",
            query="datesData",
        )

    def test_get_context_data_args(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``team.get_match_data()`` """
        self.team.get_context_data(season="2019")
        mock_get_response.assert_called_with(
            url=f"https://understat.com/team/{self.team_name}/2019",
            query="statisticsData",
        )

    def test_league_get_team_data(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``league.get_team_data()`` """
        self.league.get_team_data(season="2019")
        mock_get_response.assert_called_with(
            url=f"https://understat.com/league/{self.league_name}/2019",
            query="teamsData",
        )

    def test_league_get_match_data(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``league.get_match_data()`` """
        self.league.get_match_data(season="2019")
        mock_get_response.assert_called_with(
            url=f"https://understat.com/league/{self.league_name}/2019",
            query="datesData",
        )

    def test_league_getplayer_data(
        self, mock_get_response, mock_get, mock_request_url
    ):
        """ test ``league.get_player_data()`` """
        self.league.get_player_data(season="2019")
        mock_get_response.assert_called_with(
            url=f"https://understat.com/league/{self.league_name}/2019",
            query="playersData",
        )


@patch.object(requests.Session, "get", side_effect=mocked_requests_get)
class TestEndpointErrors(EndpointBaseTestCase):
    """Test the conditions under which exceptions are expected"""

    def test_match_get_data_bad_player(self, mock_get):
        """ test that ``match._get_data()`` raises an InvalidMatch error """
        with self.assertRaises(InvalidMatch):
            self.match.get_shot_data(status_code=404)

    def test_match_get_data_type_error(self, mock_get):
        """
        test that ``mathc.get_data()`` raises a TypeError
        when ``match`` is not a string
        """
        match = self.understat.match(match=None)
        with self.assertRaises(TypeError):
            _ = match.get_shot_data()

    def test_get_data_bad_player(self, mock_get):
        """ test that ``player._get_data()`` raises an InvalidPlayer error """
        with self.assertRaises(InvalidPlayer):
            self.player.get_shot_data(status_code=404)

    def test_player_get_data_type_error(self, mock_get):
        """
        test that ``player._get_data()`` raises a TypeError
        when ``player`` is not a string
        """
        player = self.understat.player(None)
        with self.assertRaises(TypeError):
            _ = player.get_shot_data()

    def test_team_get_data_bad_team(self, mock_get):
        """ test that ``team._get_data()`` raises an InvalidTeam error """
        team = self.understat.team(self.team_name)
        with self.assertRaises(InvalidTeam):
            _ = team.get_match_data(season="2019", status_code=404)

    def test_team_get_data_type_error(self, mock_get):
        """
        test that ``team._get_data()`` raises a TypeError
        when ``team`` is not a string
        """
        team = self.understat.team(None)
        with self.assertRaises(TypeError):
            _ = team.get_match_data(season="")

    def test_league_get_data_bad_team(self, mock_get):
        """ test that ``league._get_data()`` raises an InvalidLeague error """
        league = self.understat.league("dummy_team")
        with self.assertRaises(InvalidLeague):
            _ = league.get_match_data(season="2019", status_code=404)

    def test_league_get_data_type_error(self, mock_get):
        """
        test that ``league._get_data()`` raises a TypeError
        when ``league`` is not a string
        """
        league = self.understat.league(None)
        with self.assertRaises(TypeError):
            _ = league.get_match_data(season="2019")

    def test_invalid_season(self, mock_get):
        """
        Test that an error is raised when you try to get data for a
        season before 2014
        """
        with self.assertRaises(InvalidSeason):
            _ = self.league.get_match_data(season="2013")

    def test_error_handling_method(self, mock_get):
        # pylint: disable=no-member
        """
        test the error handling works as expected when a method is called
        that does not belong to the given endpoint
        """
        with self.assertRaises(AttributeError) as err:
            with UnderstatClient() as understat:
                understat.team("").get_bad_data()
            self.assertEqual(
                str(err),
                "'TeamEndpoint' object has no attribute 'get_bad_data'\n"
                "Its public methods are ['get_context_data', "
                "'get_match_data', get_player_data']",
            )

    @patch.object(Search, "__init__")
    def test_search_web_driver_error(self, mock_get_player_ids, mock_get):
        """
        test that ``search()`` raises a custom exception if ``geckodriver``
        is not installed
        """
        mock_get_player_ids.side_effect = WebDriverException()
        with self.assertRaisesRegex(
            WebDriverException,
            "You must have 'geckodriver' installed to use "
            "UnderstatClient.search()",
        ):
            _ = list(self.understat.search(""))


class TestEndpointDunder(EndpointBaseTestCase):
    """Tests for the dunder methods in the endpoint class"""

    def test_league(self):
        """ test ``league()`` """
        self.assertEqual(
            repr(self.understat.league(league="EPL")),
            "<LeagueEndpoint('EPL')>",
        )

    def test_player(self):
        """ test ``player()`` """
        self.assertEqual(
            repr(self.understat.player(player="1234")),
            "<PlayerEndpoint('1234')>",
        )

    def test_team(self):
        """ test ``team()`` """
        self.assertEqual(
            repr(self.understat.team(team="Manchester_United")),
            "<TeamEndpoint('Manchester_United')>",
        )

    def test_match(self):
        """ test ``match()`` """
        self.assertEqual(
            repr(self.understat.match(match="1234")), "<MatchEndpoint('1234')>"
        )

    def test_iteration(self):
        """Test iterating over players"""
        player_names = ["player_1", "player_2"]
        self.player._primary_attr = player_names
        for player, player_name in zip(self.player, player_names):
            with self.subTest(player=player_name):
                self.assertEqual(player.player, player_name)

    def test_len_one(self):
        """Test len() when there is only one player"""
        self.assertEqual(1, len(self.player))

    def test_len_error(self):
        """Test len() errors out when passed a non-sequence"""
        self.player._primary_attr = None
        with self.assertRaises(TypeError):
            self.assertEqual(1, len(self.player))

    def test_getitem_one(self):
        """Test getitem() when there is only one player"""
        self.assertEqual(self.player[0].player, self.player.player)

    def test_context_manager(self):
        """
        Test that the client behaves as a context manager as expected
        """
        try:
            with UnderstatClient():
                pass
        except Exception:  # pylint: disable=broad-except
            self.fail()


class TestSearch(unittest.TestCase):
    """ Test ``Search`` service """

    def setUp(self):
        """ setUp """
        self.url = "test/resources/home.html"
        Search.url = self.url
        Search._web_driver = MockWebDriver
        self.understat = UnderstatClient()

    def tearDown(self):
        self.understat.session.close()

    def test_search(self):
        """Test the search endpoint"""
        search = self.understat.search(player_name="Ronaldo")
        with patch(
            "understatapi.services.search.WebDriverWait", new=MockWebDriverWait
        ):
            for i, player in enumerate(search):
                # Looking at the exact output here is unhelpful because
                # of the limitations of mocking here. Need to perform
                # integration tests here
                with self.subTest(test="return value"):
                    self.assertIsInstance(player, PlayerEndpoint)
        with self.subTest(test="number of players returned"):
            self.assertEqual(i, 2)

    def test_search_with_max_ids(self):
        """
        Test the search endpoint where ``max_ids`` is less than the total
        number of results
        """
        search = self.understat.search(player_name="Ronaldo", max_ids=2)
        with patch(
            "understatapi.services.search.WebDriverWait", new=MockWebDriverWait
        ):
            for i, player in enumerate(search):
                # Looking at the exact output here is unhelpful because
                # of the limitations of mocking here. Need to perform
                # integration tests here
                with self.subTest(test="return value"):
                    self.assertIsInstance(player, PlayerEndpoint)
        with self.subTest(test="number of players returned"):
            self.assertEqual(i, 1)


if __name__ == "__main__":
    unittest.main()



""" League endpoint """
from typing import Dict, Any
import requests
from .base import BaseEndpoint
from ..exceptions import PrimaryAttribute


class LeagueEndpoint(BaseEndpoint):
    """#pylint: disable-line-too-long
    Endpoint for league data. Use this class to get data from a url of the form
    ``https://understat.com/league/<league>/<season>``

    :Example:

    .. testsetup::

        import requests
        from understatapi.endpoints import LeagueEndpoint

    .. testcleanup::

        session.close()

    .. doctest::

        >>> session = requests.Session()
        >>> leagues = ["EPL", "Bundesliga"]
        >>> for league in LeagueEndpoint(leagues, session=session):
        ...     print(league.league)
        EPL
        Bundesliga

    """

    queries = ["teamsData", "datesData", "playersData"]

    def __init__(self, league: PrimaryAttribute, session: requests.Session):
        """
        :param league: Name of the league(s) to get data for,
            one of {EPL, La_Liga, Bundesliga, Serie_A, Ligue_1, RFPL}
        :param session: The current session
        """
        self._primary_attr = league
        super().__init__(primary_attr=self._primary_attr, session=session)

    @property
    def league(self) -> PrimaryAttribute:
        """ league name """
        return self._primary_attr

    def _get_data(
        self, season: str, query: str, **kwargs: str
    ) -> Dict[str, Any]:
        """
        Get data on a league-wide basis

        :param season: Season to get data for
        :param query: Identifies the type of data to get,
            one of {teamsData, playersData, datesData}
        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        if not isinstance(self.league, str):
            raise TypeError("``league`` must be a string")
        self._check_args(league=self.league, season=season)
        url = self.base_url + "league/" + self.league + "/" + season

        data = self._get_response(url=url, query=query, **kwargs)

        return data

    def get_team_data(self, season: str, **kwargs: str) -> Dict[str, Any]:
        """
        Get data for all teams in a given league and season

        :param season: Season to get data for
        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(season=season, query="teamsData", **kwargs)
        return data

    def get_match_data(self, season: str, **kwargs: str) -> Dict[str, Any]:
        """
        Get data for all fixtures in a given league and season.

        :param season: Season to get data for
        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(season=season, query="datesData", **kwargs)
        return data

    def get_player_data(self, season: str, **kwargs: str) -> Dict[str, Any]:
        """
        Get data for all players in a given league and season

        :param season: Season to get data for
        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response()`
        """
        data = self._get_data(season=season, query="playersData", **kwargs)
        return data



""" Mock the requests library """
from requests.exceptions import HTTPError


class MockResponse:
    """ Mock response from requests.get() """

    def __init__(self, url, status_code=200, reason="OK"):
        self.url = url
        self.status_code = status_code
        self.reason = reason
        self.url = url

    @property
    def content(self):
        """ Response.content """
        with open(self.url) as file:
            content = file.read()
        return content

    @property
    def text(self):
        """ Response.content """
        with open(self.url) as file:
            text = file.read()
        return text

    def raise_for_status(self):
        """Raises ``HTTPError``, if one occurred."""

        http_error_msg = ""
        reason = self.reason
        if 400 <= self.status_code < 500:
            http_error_msg = u"%s Client Error: %s for url: %s" % (
                self.status_code,
                reason,
                self.url,
            )

        elif 500 <= self.status_code < 600:
            http_error_msg = u"%s Server Error: %s for url: %s" % (
                self.status_code,
                reason,
                self.url,
            )

        if http_error_msg:
            raise HTTPError(http_error_msg, response=self)


def mocked_requests_get(*args, **kwargs):
    """
    Return a MockResponse object
    """
    return MockResponse(*args, **kwargs)



""" Team endpoint """
from typing import Dict, Any
import requests
from requests.exceptions import HTTPError
from .base import BaseEndpoint
from ..exceptions import InvalidTeam, PrimaryAttribute


class TeamEndpoint(BaseEndpoint):
    """
    Use this class to get data from a url of the form
    ``https://understat.com/team/<team>/<season>``

    :Example:

    .. testsetup::

        import requests
        from understatapi.endpoints import TeamEndpoint

    .. testcleanup::

        session.close()

    .. doctest::

        >>> session = requests.Session()
        >>> team_names = ["Manchester_United", "Liverpool"]
        >>> for team in TeamEndpoint(team_names, session=session):
        ...     print(team.team)
        Manchester_United
        Liverpool

    """

    queries = ["datesData", "statisticsData", "playersData"]

    def __init__(
        self, team: PrimaryAttribute, session: requests.Session
    ) -> None:
        """
        :param team: Name of the team(s) to get data for
        :param session: The current session
        """
        self._primary_attr = team
        super().__init__(primary_attr=self._primary_attr, session=session)

    @property
    def team(self) -> PrimaryAttribute:
        """ team name """
        return self._primary_attr

    def _get_data(
        self, season: str, query: str, **kwargs: str
    ) -> Dict[str, Any]:
        """
        Get data on a per-team basis

        :param season: Season to get data for
        :param query: Identifies the type of data to get,
            one of {playersData, statisticsData, datesData}
        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        if not isinstance(self.team, str):
            raise TypeError("``team`` must be a string")
        self._check_args()
        url = self.base_url + "team/" + self.team + "/" + season

        try:
            data = self._get_response(url=url, query=query, **kwargs)
        except HTTPError as err:
            raise InvalidTeam(
                f"{self.team} is not a valid team", team=self.team
            ) from err

        return data

    def get_player_data(self, season: str, **kwargs: str) -> Dict[str, Any]:
        """
        Get data for all players on a given team in a given season

        :param season: Season to get data for
        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(season=season, query="playersData", **kwargs)
        return data

    def get_match_data(self, season: str, **kwargs: str) -> Dict[str, Any]:
        """
        Get data on a per match level for a given team in a given season

        :param season: Season to get data for
        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(season=season, query="datesData", **kwargs)
        return data

    def get_context_data(
        self,
        season: str,
        **kwargs: str,
    ) -> Dict[str, Any]:
        """
        Get data based on different contexts in the game

        :param season: Season to get data for
        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(season=season, query="statisticsData", **kwargs)
        return data



""" setup for understatAPI """
import os
import re
from setuptools import setup, find_packages


def read_version():
    """ read the current version from understatapi.__init__.py """
    with open(
        os.path.join(os.path.dirname(__file__), "understatapi", "__init__.py"),
        encoding="utf-8",
    ) as fpath:
        for line in fpath:
            match = re.search(
                r'^\s*__version__\s*=\s*([\'"])([^\'"]+)\1\s*$', line
            )
            if match:
                return match.group(2)
    raise RuntimeError("Unable to find version")


def long_description():
    """ Get description from readme """
    with open("README.md", "r", encoding="utf-8") as fpath:
        long_des = fpath.read()
    return long_des


setup(
    name="understatapi",
    version=read_version(),
    description="An API for scraping data from understat.com",
    packages=find_packages(exclude=("tests",)),
    long_description=long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/collinb9/understatAPI",
    author="collinb9",
    author_email="brendan.m.collins@outlook.com",
    license="MIT",
    install_requires=["requests>=2.0.0", "selenium>=3.141.0"],
    keywords=(
        "statistics xG expected goals fpl fantasy"
        "premier league understat football web scraping"
        "scraper"
    ),
    project_urls={
        "Source": "https://github.com/collinb9/understatAPI/",
    },
)



""" Base endpoint """
from typing import List, Sequence, Dict, Any
import json
import requests
from requests import Response
from ..exceptions import (
    InvalidLeague,
    InvalidSeason,
    PrimaryAttribute,
)


class BaseEndpoint:
    """
    Base endpoint for understat API

    :attr base_url: str: The base url to use for requests,
        ``https://understat.com/``
    :attr leagues: List[str]: The available leagues, ``EPL``, ``La_Liga``,
        ``Bundesliga``, optional``Serie_A``, ``Ligue_1``, ``RFPL``
    :attr queries: List[str]: Strings that can be searched for in the html
        pages.
    """

    base_url = "https://understat.com/"
    leagues = ["EPL", "La_Liga", "Bundesliga", "Serie_A", "Ligue_1", "RFPL"]
    queries: List[str] = []

    def __init__(
        self, primary_attr: PrimaryAttribute, session: requests.Session
    ) -> None:
        """
        :session: requests.Session: The current ``request`` session
        """
        self.session = session
        self._primary_attr = primary_attr

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self._primary_attr!r})>"

    def __len__(self) -> int:
        if isinstance(self._primary_attr, str):
            return 1
        if isinstance(self._primary_attr, Sequence):
            return len(self._primary_attr)
        raise TypeError("Primary attribute is not a sequence or string")

    def __getitem__(self, index: int) -> "BaseEndpoint":
        if index >= len(self):
            raise IndexError
        if isinstance(self._primary_attr, str):
            return self.__class__(self._primary_attr, session=self.session)
        return self.__class__(self._primary_attr[index], session=self.session)

    def _check_args(self, league: str = None, season: str = None) -> None:
        """ Handle invalid arguments """
        if league is not None and league not in self.leagues:
            raise InvalidLeague(
                f"{league}is not a valid league", league=league
            )
        if season is not None and int(season) < 2014:
            raise InvalidSeason(
                f"{season} is not a valid season", season=season
            )

    def _request_url(self, *args: str, **kwargs: str) -> Response:
        """
        Use the requests module to send a HTTP request to a url, and check
        that this request worked.

        :param args: Arguments to pass to ``requests.get()``
        :param kwargs: Keyword arguments to pass to ``requests.get()``
        """
        res = self.session.get(*args, **kwargs)
        res.raise_for_status()
        return res

    @staticmethod
    def _extract_data_from_html(
        html: str,
        query: str = "teamsData",
    ) -> Dict[str, Any]:
        """
        Finds a JSON in the HTML according to a query, and returns the
        dictionary corresponding to this JSON.

        :param html: A html document
        :param query: A sub-string to look for in the html document
        """
        # find the query in the html string
        query_index = html.find(query)
        # get the start and end of the JSON data string
        start = html.find("(", query_index) + 2
        end = html.find(")", start) - 1
        json_data = html[start:end]
        # Clean up the json and return the data
        json_data = json_data.encode("utf8").decode("unicode_escape")
        data = json.loads(json_data)
        return data

    def _get_response(
        self,
        url: str,
        query: str = "teamsData",
        **kwargs: str,
    ) -> Dict[str, Any]:
        """
        Retrieve data from html page

        :param url: str: url to parse
        :param query: str: A sub-string to look for in the html document
        :param kwargs: Keyword arguments to pass to ``requests.get()``

        :return: JSON data retrieved from html page
        """
        res = self._request_url(url, **kwargs)
        data = self._extract_data_from_html(res.text, query=query)

        return data



""" Integration tests fo understatapi """
import os
import unittest
import tracemalloc
import requests
from understatapi.services import Search
from understatapi import UnderstatClient

try:
    RUN_TESTS = bool(os.environ["UNDERSTAT_INTEGRATION"])
except KeyError:
    RUN_TESTS = False


@unittest.skipIf(not RUN_TESTS, "UNDERSTAT_INTEGRATION not set to True")
class TestIntegrationSeleium(unittest.TestCase):
    """ Integration test for functions using selenium """

    def setUp(self):
        self.search = Search(player_name="Ronaldo", session=requests.Session())

    def tearDown(self):
        self.search.__exit__(None, None, None)
        self.search.session.close()

    def test_get_player_ids(self):
        """ test ``get_player_ids()`` """
        self.assertListEqual(
            list(self.search.get_player_ids()), ["2371", "2028", "7097"]
        )

    def test_search_invalid_player(self):
        """
        test what happens when ``Search`` is passed a search term which has
        no matches
        """
        self.search.player_name = "abcdef"
        self.assertListEqual(list(self.search.get_player_ids()), [])


@unittest.skipIf(not RUN_TESTS, "UNDERSTAT_INTEGRATION not set to True")
class TestIntegrationUnderstatClient(unittest.TestCase):
    """ Integration tests for interactions with ``UnderstatClient`` """

    def setUp(self):
        self.understat = UnderstatClient()

    def tearDown(self):
        self.understat.__exit__(None, None, None)

    def test_cookie_transfer(self):
        """
        Test that cookies get transferred from the selenium ``WebDriver``
        object to the ``requests.Session`` object
        """
        with self.subTest(test="before"):
            self.assertEqual(len(self.understat.session.cookies), 0)
        list(self.understat.search("Cristiano Ronaldo"))
        with self.subTest(test="after"):
            self.assertEqual(len(self.understat.session.cookies), 7)


if __name__ == "__main__":
    tracemalloc.start()
    unittest.main()



""" Match endpoint """
from typing import Dict, Any
import requests
from requests.exceptions import HTTPError
from .base import BaseEndpoint
from ..exceptions import InvalidMatch, PrimaryAttribute


class MatchEndpoint(BaseEndpoint):
    """
    Use this class to get data from a url of the form
    ``https://understat.com/match/<match_id>``

    :Example:

    .. testsetup::

        import requests
        from understatapi.endpoints import MatchEndpoint

    .. testcleanup::

        session.close()

    .. doctest::

        >>> session = requests.Session()
        >>> match_ids = ["123", "456"]
        >>> for match in MatchEndpoint(match_ids, session=session):
        ...     print(match.match)
        123
        456
    """

    queries = ["shotsData", "rostersData", "match_info"]

    def __init__(self, match: PrimaryAttribute, session: requests.Session):
        """
        :param match: Id of match(es) to get data for
        :param session: The current session
        """
        self._primary_attr = match
        super().__init__(primary_attr=self._primary_attr, session=session)

    @property
    def match(self) -> PrimaryAttribute:
        """  match id """
        return self._primary_attr

    def _get_data(self, query: str, **kwargs: str) -> Dict[str, Any]:
        """
        Get data on a per-match basis

        :param query: Identifies the type of data to get,
            one of {shotsData, rostersData, match_info}
        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        if not isinstance(self.match, str):
            raise TypeError("``match`` must be a string")
        self._check_args()
        url = self.base_url + "match/" + self.match

        try:
            data = self._get_response(url=url, query=query, **kwargs)
        except HTTPError as err:
            raise InvalidMatch(
                f"{self.match} is not a valid match", match=self.match
            ) from err

        return data

    def get_shot_data(self, **kwargs: str) -> Dict[str, Any]:
        """
        Get shot level data for a match

        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(query="shotsData", **kwargs)
        return data

    def get_roster_data(self, **kwargs: str) -> Dict[str, Any]:
        """
        Get data about the roster for each team

        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(query="rostersData", **kwargs)
        return data

    def get_match_info(self, **kwargs: str) -> Dict[str, Any]:
        """
        Get information about the match

        :param kwargs: Keyword argument to pass to
            :meth:`understatapi.endpoints.base.BaseEndpoint._get_response`
        """
        data = self._get_data(query="match_info", **kwargs)
        return data
